# Computes Cornu spline that passes through a given set of points
#
# A ruby adaptation [for use with SketchUp]
# of path.py created by Raph Levien ( www.levien.com )
# 
# path.py is available at http://casper.ghostscript.com/~raph/cornu/
# also see http://levien.com/spiro/ for additional information
#
# Insights provided by Cornu for Nodebox 
# available at http://nodebox.net/code/index.php/Cornu
#
# This code is licensed under the GNU General Public License, version 2 or later
#
# The underlying curve technology is subject to a patent
# 
# Ruby adaptation by Terry Ross ( www.drawmetal.com )

class Cornu
  
  include Math  
  def initialize

    @fres_integ = FresnelIntegral.new
    return
  end 
# ----------------------------------------------------------------------------------------------
  
  def mod_2pi(th)
    
    th = -1.0e12 if th <= -1.0e12
    th = 1.0e12 if th >= 1.0e12
    u = th / (2 * PI)
    return 2 * PI * (u - (u + 0.5).floor)

  end
  
  def fit_cornu_half(th0, th1)
    if th0 + th1 < 1e-6
      epsilon = 1e-6
      th0 += epsilon
      th1 += epsilon
    end
    n_iter = 0
    n_iter_max = 21
    est_tm = 0.29112 * (th1 + th0) / sqrt(th1 - th0)
    l = est_tm * 0.9
    r = est_tm * 2
    while true
      t_m = 0.5 * (l + r)
      dt = (th0 + th1) / (4 * t_m)
      t0 = t_m - dt
      t1 = t_m + dt

      # invariant: t1^2 - t0^2 = th0 + th1
      s0, c0 = @fres_integ.eval_cornu(t0)
      s1, c1 = @fres_integ.eval_cornu(t1)

      chord_th = atan2(s1 - s0, c1 - c0)
      n_iter += 1
      if n_iter == n_iter_max
          break
      end
      if mod_2pi(chord_th - t0 * t0 - th0) < 0
          l = t_m
      else
          r = t_m
      end
    end
    chordlen = hypot(s1 - s0, c1 - c0)
    k0 = t0 * chordlen
    k1 = t1 * chordlen
    return t0, t1, k0, k1
  end

  # fit arc to pts (0, 0), (x, y), and (1, 0), return th tangent to
  # arc at (x, y)
  def fit_arc(x, y)
    th = atan2(y - 2 * x * y, y * y + x - x * x) 
    return th
  end
  
  # find thetas tangent to path using local cspline logic
  def local_ths(path, closed)
    if closed
      result = []
      n_closed = 1
    else
      result = [0]
      n_closed = 0
    end
    for i in (1 - n_closed...path.length - 1 + n_closed)
      x0, y0 = path[(i + path.length - 1) % path.length]
      x1, y1 = path[i]
      x2, y2 = path[(i + 1) % path.length]
      dx = x2 - x0
      dy = y2 - y0
      ir2 = dx * dx + dy * dy
      x = ((x1 - x0) * dx + (y1 - y0) * dy) / ir2
      y = ((y1 - y0) * dx - (x1 - x0) * dy) / ir2
      th = fit_arc(x, y) + atan2(dy, dx)
      result.push(th)
    end
    if not closed
      result.push(0)
    end
    boundary_ths(path, result, closed)
    return result
  end
  
  # set the endpoint thetas so endpoint curves are circular arcs
  def boundary_ths(path, ths, closed)
    if not closed
      first_th = 2 * atan2(path[1][1] - path[0][1], path[1][0] - path[0][0]) - ths[1]
      ths[0] = first_th
      last_th = 2 * atan2(path[-1][1] - path[-2][1], path[-1][0] - path[-2][0]) - ths[-2]
      ths[-1] = last_th
    end
  end
     
  def draw_cornu(path, ths, closed, sides)

    side_factor = 1.0 / sides.to_f
    n_closed = 0
    n_closed = 1 if closed
    
    for i in (0...path.length - 1 + n_closed)
      x0, y0 = path[i]
      x1, y1 = path[(i + 1) % path.length]
      th = atan2(y1 - y0, x1 - x0)
      th0 = mod_2pi(ths[i] - th)
      th1 = mod_2pi(th - ths[(i + 1) % path.length])
      flip = -1
      th1 += 1e-6
      if th1 < th0
          th0, th1 = th1, th0
          flip = 1
      end
      t0, t1, k0, k1 = fit_cornu_half(th0, th1)
      if flip == 1
          t0, t1 = t1, t0
          k0, k1 = k1, k0
      end    
      s0, c0 = @fres_integ.eval_cornu(t0)
      s0 *= flip
      s1, c1 = @fres_integ.eval_cornu(t1)
      s1 *= flip

      chord_th = atan2(s1 - s0, c1 - c0)
      chordlen = hypot(s1 - s0, c1 - c0)
      rot = th - chord_th
      scale = hypot(y1 - y0, x1 - x0) / chordlen
      cs = scale * cos(rot)
      ss = scale * sin(rot)
  
      for j in (0...sides)
        t = j * side_factor
        s, c = @fres_integ.eval_cornu(t0 + t * (t1 - t0))
        s *= flip
        s -= s0
        c -= c0
        x = c * cs - s * ss
        y = s * cs + c * ss
        print_pt(x0 + x, y0 + y)
      end
    end
    
    if closed
      # add the first point in the path to the curve (again)
      index = 0
    else
      # add the last point in the path to the curve
      index = -1
    end
    x0, y0 = path[index]
    print_pt(x0, y0)
    return

  end
        
  # update thetas based on cornu splines
  def tweak_ths(path, ths, closed)
    n_closed = 0
    n_closed = 1 if closed
    last_k1 = 0
    
    dks = []
    for i in (0...path.length - 1 + n_closed)
      x0, y0 = path[i]
      x1, y1 = path[(i + 1) % path.length]
      th = atan2(y1 - y0, x1 - x0)
      th0 = mod_2pi(ths[i] - th)
      th1 = mod_2pi(th - ths[(i + 1) % path.length])
      flip = -1
      th1 += 1e-6
      if th1 < th0
        th0, th1 = th1, th0
        flip = 1
      end
      t0, t1, k0, k1 = fit_cornu_half(th0, th1)
      if flip == 1
        t0, t1 = t1, t0
        k0, k1 = k1, k0
      end
      s0, c0 = @fres_integ.eval_cornu(t0)
      s1, c1 = @fres_integ.eval_cornu(t1)
      chordlen = hypot(s1 - s0, c1 - c0)
      scale = 1 / hypot(y1 - y0, x1 - x0)
      k0 *= scale
      k1 *= scale

      if i > 0 
        dk = k0 - last_k1
        dks.push(dk)
      else
        first_k0 = k0
      end
      last_k1 = k1

    end
    
    if closed
      dks.push(first_k0 - last_k1)
    end
    
    for i in (0...dks.length)
      x0, y0 = path[i]
      x1, y1 = path[(i + 1) % path.length]
      x2, y2 = path[(i + 2) % path.length]

      chord1 = hypot(x1 - x0, y1 - y0)
      chord2 = hypot(x2 - x1, y2 - y1)
      ths[(i + 1) % path.length] -= 0.5 * (dks[i] / (chord1 + chord2))
    end
  end 
  
  def print_pt(x, y)
    @out_path[@p] = [x, y]
    @p += 1
    return
  end
  
  def drawpath(in_path, closed, iterations, sides)
# in_path is an array of 2D points through which the spline will be fit
# closed = true to create a spline that connects last point back to first point
# iterations controls effort expended by fitting algorithm (larger = more effort)
# sides is number of staight line segments in spline point to point along in_path
# out_path is 2D Cornu spline that passed through in_path

    path = in_path.dup
    ths = local_ths(path, closed)
    
    t_start = Time.new
    t_last = t_start
    time_msg = false
    
    for i in (0...iterations)
      
        boundary_ths(path, ths, closed)
        tweak_ths(path, ths, closed)
        
        t_now = Time.new
        forecast = (t_now - t_start) * (iterations / (i+1) )
        
        if time_msg == true
          since_last = t_now - t_last 
          if since_last > 3.0
            actual += since_last
            remain = forecast - actual
            remain = 2.0 if remain <= 2.0
            remain = CTSUtils.round_value( remain, 0 )
            Sketchup::set_status_text $ctst.to( "Estimated time remaining:" ) + " " + remain.to_s + " " + $ctst.to( "seconds" )
            t_last = t_now
          end
        end

        if forecast > 3.0
          if time_msg == false
            time_msg = true
            actual = t_now - t_last 
            forecast = CTSUtils.round_value( forecast, 0 )
            Sketchup::set_status_text $ctst.to( "Estimated time to draw:" ) + " " +  forecast.to_s + " " +  $ctst.to( "seconds" )
            t_last = t_now
          end     
        end        
        
    end
    
    boundary_ths(path, ths, closed)
    
    @out_path = []    
    @p = 0
    draw_cornu(path, ths, closed, sides)

    return @out_path

  end

end


class FresnelIntegral
  include Math
  
  def initialize

    @sn = [
    -2.99181919401019853726E3,
     7.08840045257738576863E5,
    -6.29741486205862506537E7,
     2.54890880573376359104E9,
    -4.42979518059697779103E10,
     3.18016297876567817986E11
    ]
    @sn.reverse!
    @sd = [
     1.00000000000000000000E0,
     2.81376268889994315696E2,
     4.55847810806532581675E4,
     5.17343888770096400730E6,
     4.19320245898111231129E8,
     2.24411795645340920940E10,
     6.07366389490084639049E11
    ]
    @sd.reverse!
    @cn = [
    -4.98843114573573548651E-8,
     9.50428062829859605134E-6,
    -6.45191435683965050962E-4,
     1.88843319396703850064E-2,
    -2.05525900955013891793E-1,
     9.99999999999999998822E-1
    ]
    @cn.reverse!
    @cd = [
     3.99982968972495980367E-12,
     9.15439215774657478799E-10,
     1.25001862479598821474E-7,
     1.22262789024179030997E-5,
     8.68029542941784300606E-4,
     4.12142090722199792936E-2,
     1.00000000000000000118E0
    ]
    @cd.reverse!
    @fn = [
      4.21543555043677546506E-1,
      1.43407919780758885261E-1,
      1.15220955073585758835E-2,
      3.45017939782574027900E-4,
      4.63613749287867322088E-6,
      3.05568983790257605827E-8,
      1.02304514164907233465E-10,
      1.72010743268161828879E-13,
      1.34283276233062758925E-16,
      3.76329711269987889006E-20
    ]
    @fn.reverse!
    @fd = [
      1.00000000000000000000E0,
      7.51586398353378947175E-1,
      1.16888925859191382142E-1,
      6.44051526508858611005E-3,
      1.55934409164153020873E-4,
      1.84627567348930545870E-6,
      1.12699224763999035261E-8,
      3.60140029589371370404E-11,
      5.88754533621578410010E-14,
      4.52001434074129701496E-17,
      1.25443237090011264384E-20
    ]
    @fd.reverse!
    @gn = [
      5.04442073643383265887E-1,
      1.97102833525523411709E-1,
      1.87648584092575249293E-2,
      6.84079380915393090172E-4,
      1.15138826111884280931E-5,
      9.82852443688422223854E-8,
      4.45344415861750144738E-10,
      1.08268041139020870318E-12,
      1.37555460633261799868E-15,
      8.36354435630677421531E-19,
      1.86958710162783235106E-22
    ]
    @gn.reverse!
    @gd = [
      1.00000000000000000000E0,
      1.47495759925128324529E0,
      3.37748989120019970451E-1,
      2.53603741420338795122E-2,
      8.14679107184306179049E-4,
      1.27545075667729118702E-5,
      1.04314589657571990585E-7,
      4.60680728146520428211E-10,
      1.10273215066240270757E-12,
      1.38796531259578871258E-15,
      8.39158816283118707363E-19,
      1.86958710162783236342E-22
    ]
    @gd.reverse!
  
    return
  end

  # implementation of computation of fresnel integrals adapted from Cephes Math Library
  
  def polevl(x, coef)
    ans = coef[-1] 
    #for i in range(len(coef) - 2, -1, -1):
    i = coef.length - 2
    while i > -1
      ans = ans * x + coef[i]
      i += -1
    end
    return ans
  end   
  
  def fresnel(xxa)
    x = xxa.abs
    x2 = x * x
    if x2 < 2.5625
      t = x2 * x2
      ss = x * x2 * polevl(t, @sn) / polevl(t, @sd)
      cc = x * polevl(t, @cn) / polevl(t, @cd)
    elsif x > 36974.0
      ss = 0.5
      cc = 0.5
    else
      t = PI * x2
      u = 1.0 / (t * t)
      t = 1.0 / t
      f = 1.0 - u * polevl(u, @fn) / polevl(u, @fd)
      g = t * polevl(u, @gn) / polevl(u, @gd)
      t = PI * 0.5 * x2
      c = cos(t)
      s = sin(t)
      t = PI * x
      cc = 0.5 + (f * s - g * c) / t
      ss = 0.5 - (f * c + g * s) / t
    end
    if xxa < 0
      cc = -cc
      ss = -ss
    end
    return ss, cc
  end

# end of adaptation from Cephes Math Library

  def eval_cornu(t)
    spio2 = sqrt(PI * 0.5)
    s, c = fresnel(t / spio2)
    s *= spio2
    c *= spio2
    return s, c
  end

end

